#include "demo.h"

#define FCY		40000000
#define BR_USB		115200
#define BRG_USB		((FCY / BR_USB) / 16) - 1

#define	RX_BUF_SIZE	128
#define	TX_BUF_SIZE	128

uint8_t	rx1Buf[RX_BUF_SIZE];
uint8_t	tx1Buf[TX_BUF_SIZE];
queue_t	rx1Queue;
queue_t	tx1Queue;

/*
 * Initialize the uart for 115200 baud. This implementation is interrupt driven.
 * To transmiit characters you enter them into the tx queue and let the ISR take them
 * out. For receive, the ISR inserts them into the queue and the software take them out.
 * Queue size is not terribly important for this project.
 */
void uartInit(void)
{
    /*
     * Initialize the data structures
     */
    queueInit(&rx1Queue, rx1Buf, RX_BUF_SIZE);
    queueInit(&tx1Queue, tx1Buf, TX_BUF_SIZE);
    
    /*
     * Map the I/O pins
     */
    RPINR18bits.U1RXR = 3;
    RPOR8bits.RP16R = RPO_U1TX;
	
    /*
     * Setup the interrupts
     */
    IPC2bits.U1RXIP = 1;			// set UART 1 RX interrupt priority level
    IFS0bits.U1RXIF = 0;			// clear RX interrupt flag
    IEC0bits.U1RXIE = 1;			// enable RX interrupts

    IPC3bits.U1TXIP = 1;			// set UART 1 1X interrupt priority level
    IFS0bits.U1TXIF = 0;			// clear TX interrupt flag
    IEC0bits.U1TXIE = 1;			// enable TX interrupts

    U1BRG = BRG_USB;				// baud rate divisor

    U1MODE = (1 << 15);
    U1STAbits.UTXEN = 1;			// tx enable
}

/*
 * UART 1 RX ISR
 */
void __attribute__((__interrupt__, no_auto_psv)) _U1RXInterrupt(void)
{
    uint8_t
	ch;

    ch = U1RXREG;
    queueInsert(&rx1Queue, ch);
    IFS0bits.U1RXIF = 0;			// clear interrupt flag
}

/*
 * UART 1 TX ISR
 */
void __attribute__((__interrupt__, no_auto_psv)) _U1TXInterrupt(void)
{
    uint8_t
	ch;
    int
	rc;

    rc = queueExtract(&tx1Queue, &ch);
    if (rc == 0)
    {
	IEC0bits.U1TXIE = 0;
    }
    else
    {
	IFS0bits.U1TXIF = 0;			// clear interrupt flag
	U1TXREG = ch;
    }
}

/*
 * Display one character on the console. Blocks until there is room in the queue.
 */
void uartPutc(uint8_t txd)
{
//    if (txd == '\n') uartPutc('\r');
    while (queueInsert(&tx1Queue, txd) == 0) Nop();
    _U1TXIE = 1;				// enable TX interrupts
}

/*
 * This is the "C" library version. Translate the newline into a carriage return
 * new line sequence. This would't be very good for data transfer but works for printf!
 */
int putc(int ch, FILE *fd)
{
    uartPutc(ch);
    if (ch == '\n') uartPutc('\r');
    return(1);
}

/*
 * Extract a single character fro the receive queue. Blocks until a char is ready.
 */
uint8_t uartGetc(void)
{
    uint8_t
	rxd;

    while (queueExtract(&rx1Queue, &rxd) == 0) Nop();
    return(rxd);
}

int getc(FILE *fd)
{
    return(uartGetc());
}

/*
 * Test if there are any character pending in the rx queue.
 */
int uartCharAvail(void)
{
    return(rx1Queue.bcnt != 0);
}

/*
 * Get a character string from the console. Do primitive editting.
 */
char *uartGets(char *s)
{
    char
	*p, ch;
    int
	n;

    n = 0;
    p = s;
    do
    {
	ch = uartGetc();
	switch (ch)
	{
	    case '\r':
		*p = 0;
		uartPutc('\r');
		uartPutc('\n');
		break;
	    case '\b':
		if (n > 0)
		{
		    uartPutc('\b');
		    uartPutc(' ');
		    uartPutc('\b');
		    n--;
		    p--;
		}
		break;
	    default:
		if ((ch >= ' ') && (ch <= 'z'))
	        {
		    uartPutc(ch);
		    n++;
		    *p++ = ch;
	        }
		break;
	}
    }
    while (ch != '\r');
    return(s);
}
